/*
 *  i386 Host code generation
 * 
 *  Copyright (c) 2005 CodeSourcery, LLC
 *  Written by Paul Brook
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define HOST_ENV_REG HOST_REG_AREG0
static inline int get_arg_regno(int qreg)
{
    qemu_assert(qreg >= FIRST_HARD_QREG, "bad hard regno");
    return qreg - FIRST_HARD_QREG;
}

static inline long get_arg_im(int qreg)
{
    qemu_assert(gen_reginfo[qreg].type == QREGT_IM, "bad immediate qreg");
    return gen_reginfo[qreg].val;
}

#define arg_regno(n) get_arg_regno(ARG(n))
#define arg_im(n) get_arg_im(ARG(n))

#define DO_FIXUP_PCREL(from, to) \
    *(uint32_t *)from -= to;

/* Encode a Mod R/M byte for a register operand.  */
#define MODRM(reg, rm) OUT8(0xc0 | ((reg) << 3) | (rm))

/* Encode a Mod R/M byte and offset for a base+offsed memory address.  */
#define MODRM_OFFSET(reg, base, offset) do { \
    if ((int8_t)(offset) == (int32_t)(offset)) { \
        if (base == HOST_REG_ESP) { \
            OUT8(0x44 | ((reg) << 3)); \
            OUT8(0x24); \
        } else { \
            OUT8(0x40 | ((reg) << 3) | (base)); \
        } \
        OUT8(offset); \
    } else { \
        if (base == HOST_REG_ESP) { \
            OUT8(0x84 | ((reg) << 3)); \
            OUT8(0x24); \
        } else { \
            OUT8(0x80 | ((reg) << 3) | (base)); \
        } \
        OUT32(offset); \
    }}while(0)

QOP_CODE_START
QOP_START(mov32)
    int dest_qreg = ARG(0);
    int src_qreg = ARG(1);
    struct qreg *info;
    int regno;
    int qreg;

    qreg = QREG_NULL;
    if (src_qreg < FIRST_HARD_QREG) {
        info = &gen_reginfo[src_qreg];
        regno = arg_regno(0);
        if (info->type == QREGT_IM) {
            OUT8(0xb8 + regno); OUT32(info->val);
        } else {
            qreg = src_qreg;
        }
    } else if (dest_qreg < FIRST_HARD_QREG) {
        qreg = dest_qreg;
        info = &gen_reginfo[dest_qreg];
        regno = arg_regno(1);
    } else {
        OUT8(0x89); MODRM(arg_regno(1), arg_regno(0));
    }
    if (qreg) {
        int val;
        val = info->val;
        if (qreg == src_qreg)
            OUT8(0x8b);
        else
            OUT8(0x89);
        switch (info->type) {
        case QREGT_STACK:
            MODRM_OFFSET(regno, HOST_REG_ESP, val);
            break;
        case QREGT_ENV:
            MODRM_OFFSET(regno, HOST_ENV_REG, val);
            break;
        default:
            cpu_abort(NULL, "host-i386: Bad qreg type"); \
        }
    }
QOP_END

#define MODRM_FLOAT64(reg, tmp_reg, high) do { \
    uint32_t offset; \
    offset = high ? 4 : 0; \
    if (reg >= FIRST_HARD_QREG) { \
        offset += offsetof(CPUState, float64regs[get_arg_regno(reg) \
                                                 - FIRST_FLOAT64_REG]); \
        MODRM_OFFSET(tmp_reg, HOST_ENV_REG, offset); \
    } else { \
        struct qreg *info; \
        info = &gen_reginfo[reg]; \
        offset += info->val; \
        switch (info->type) { \
        case QREGT_ENV: \
            MODRM_OFFSET(tmp_reg, HOST_ENV_REG, offset); \
            break; \
        case QREGT_STACK: \
            MODRM_OFFSET(tmp_reg, HOST_REG_ESP, offset); \
            break; \
        default: \
            cpu_abort(NULL, "host-i386: Bad qreg type"); \
        } \
    }}while(0)

QOP_START(movf64)
    int dest_qreg = ARG(0);
    int src_qreg = ARG(1);
    /* The register allocator can't cope with allocating a scratch for a
       simple move.  Use floating point instructions and hope that doesn't
       confuse anything else.  */
    /* fldl */
    OUT8(0xdd); MODRM_FLOAT64(src_qreg, 0, 0);
    /* fstpl */
    OUT8(0xdd); MODRM_FLOAT64(dest_qreg, 3, 0);
QOP_END

QOP_START(zerof64)
    long offset;

    offset = offsetof(CPUState, float64regs[arg_regno(0)
                                            - FIRST_FLOAT64_REG]);
    /* movl $0, val  */
    OUT8(0xc7); MODRM_OFFSET(0, HOST_ENV_REG, offset); OUT32(0);
    OUT8(0xc7); MODRM_OFFSET(0, HOST_ENV_REG, offset + 4); OUT32(0);
QOP_END

QOP_START(neg32)
    OUT8(0xf7); MODRM(3, arg_regno(0));
QOP_END

QOP_START(not32)
    OUT8(0xf7); MODRM(2, arg_regno(0));
QOP_END

QOP_START(bswap32)
    OUT8(0x0f); OUT8(0xc8 + arg_regno(0));
QOP_END

QOP_START(and32)
    OUT8(0x23); MODRM(arg_regno(0), arg_regno(2));
QOP_END

QOP_START(or32)
    OUT8(0x0b); MODRM(arg_regno(0), arg_regno(2));
QOP_END

QOP_START(xor32)
    OUT8(0x33); MODRM(arg_regno(0), arg_regno(2));
QOP_END

QOP_START(add32)
    OUT8(0x03); MODRM(arg_regno(0), arg_regno(2));
QOP_END

QOP_START(sub32)
    OUT8(0x2b); MODRM(arg_regno(0), arg_regno(2));
QOP_END

#define DO_SHIFT(name, op) \
QOP_START(name ## 32) \
    if (ARG(2) >= FIRST_HARD_QREG) { \
        qemu_assert(arg_regno(2) == HOST_REG_ECX, "bad shift qop"); \
        OUT8(0xd3); MODRM(op, arg_regno(0)); \
    } else { \
        int n; \
        n = arg_im(2); \
        if (n == 1) \
            OUT8(0xd1); \
        else \
            OUT8(0xc1); \
        MODRM(op, arg_regno(0)); \
        if (n != 1) \
            OUT8(n); \
    } \
QOP_END

DO_SHIFT(shr, 5)
DO_SHIFT(sar, 7)
DO_SHIFT(shl, 4)
DO_SHIFT(ror, 1)
DO_SHIFT(rol, 0)

QOP_START(mul32)
    int dest = arg_regno(0);
    int src = arg_regno(2);
    int scratch = arg_regno(3);

    qemu_assert (dest == HOST_REG_EAX && scratch == HOST_REG_EDX,
                 "invalid mul qop");
    OUT8(0xf7); MODRM(4, src);
QOP_END

#define DO_EXT(name, op) \
QOP_START(ext ## name ## 32) \
    OUT8(0x0f); OUT8(op); MODRM(arg_regno(0), arg_regno(1)); \
QOP_END
DO_EXT(8u, 0xb6)
DO_EXT(16u, 0xb7)
DO_EXT(8s, 0xbe)
DO_EXT(16s, 0xbf)

#define DO_SETcc32(name, cc) \
QOP_START(set_##name##32) \
    int dest = arg_regno(0); \
    int src1 = arg_regno(1); \
    int src2 = arg_regno(2); \
    /* cmpl */ \
    OUT8(0x3b); MODRM(src1, src2); \
    /* set<cc> dest */ \
    OUT8(0x0f); OUT8(0x90 + cc); MODRM(0, dest); \
    /* movzbl dest, dest */ \
    OUT8(0x0f); OUT8(0xb6); MODRM(dest, dest); \
QOP_END
DO_SETcc32(ltu, 2)
DO_SETcc32(leu, 6)

#define DO_JMPcc32(name, op) \
QOP_START(jmp_##name##32) \
    int regno = arg_regno(0); \
    /* testl src, src */ \
    OUT8(0x85); MODRM(regno, regno); \
    /* j<cc> dest */ \
    OUT8(0x0f); OUT8(0x80 + op); \
    OUT32(LABEL_OFFSET(ARG(1)) - 4); \
QOP_END

DO_JMPcc32(z, 4)
DO_JMPcc32(nz, 5)
DO_JMPcc32(s, 8)
DO_JMPcc32(ns, 9)

QOP_START(jmp)
    OUT8(0xe9); OUT32(LABEL_OFFSET(ARG(0)) - 4);
QOP_END

QOP_START(helper32)
    long fixup;

    /* We should have input in EDX and output in EAX.  */
    qemu_assert (arg_regno(0) == HOST_REG_EAX, "bad reg in helper");
    qemu_assert (arg_regno(1) == HOST_REG_EDX, "bad reg in helper");

#ifdef HOST_USE_REGPARM
    /* movl %ebp, %eax */
    OUT8(0x8b); MODRM(HOST_REG_EAX, HOST_ENV_REG);
#else
    /* pushl src */
    OUT8(0x50 + HOST_REG_EDX);
    /* pushl %ebp */
    OUT8(0x50 + HOST_ENV_REG);
#endif
    /* call helper */
    OUT8(0xe8); MARK_FIXUP(fixup); OUT32((uint32_t)gen_helper_address[ARG(2)]);
    RESOLVE_FIXUP(PCREL, fixup);

#ifndef HOST_USE_REGPARM
    /* addl $8, %esp */
    OUT8(0x83); MODRM(0, HOST_REG_ESP); OUT8(8);
#endif
QOP_END

QOP_START(helper32_32)
    long fixup;

    /* We should have input in EDX, ECX and output in EAX.  */
    qemu_assert (arg_regno(0) == HOST_REG_EAX, "bad reg in helper");
    qemu_assert (arg_regno(1) == HOST_REG_EDX, "bad reg in helper");
    qemu_assert (arg_regno(2) == HOST_REG_ECX, "bad reg in helper");

#ifdef HOST_USE_REGPARM
    /* movl %ebp, %eax */
    OUT8(0x8b); MODRM(HOST_REG_EAX, HOST_ENV_REG);
#else
    /* pushl arg2 */
    OUT8(0x50 + HOST_REG_ECX);
    /* pushl arg1 */
    OUT8(0x50 + HOST_REG_EDX);
    /* pushl %ebp */
    OUT8(0x50 + HOST_ENV_REG);
#endif
    /* call helper */
    OUT8(0xe8); MARK_FIXUP(fixup); OUT32((uint32_t)gen_helper_address[ARG(3)]);
    RESOLVE_FIXUP(PCREL, fixup);

#ifndef HOST_USE_REGPARM
    /* addl $12, %esp */
    OUT8(0x83); MODRM(0, HOST_REG_ESP); OUT8(12);
#endif
QOP_END

QOP_START(helperf64x32)
    long fixup;
    qemu_assert (arg_regno(0) == HOST_REG_EAX, "bad reg in helper");
#ifdef HOST_USE_REGPARM
    /* movl %ebp, %eax */
    OUT8(0x8b); MODRM(HOST_REG_EAX, HOST_ENV_REG);
    /* movl $arg1, %edx */
    OUT8(0xb8 + HOST_REG_EDX); OUT32(arg_regno(1) - FIRST_FLOAT64_REG);
#else
#error not implemented
#endif
    /* call helper */
    OUT8(0xe8); MARK_FIXUP(fixup); OUT32((uint32_t)gen_helper_address[ARG(2)]);
    RESOLVE_FIXUP(PCREL, fixup);
QOP_END

QOP_START(helper32xf64)
    long fixup;
    qemu_assert (arg_regno(1) == HOST_REG_ECX, "bad reg in helper");
#ifdef HOST_USE_REGPARM
    /* movl %ebp, %eax */
    OUT8(0x8b); MODRM(HOST_REG_EAX, HOST_ENV_REG);
    /* movl $result, %edx */
    OUT8(0xb8 + HOST_REG_EDX); OUT32(arg_regno(0) - FIRST_FLOAT64_REG);
#else
#error not implemented
#endif
    /* call helper */
    OUT8(0xe8); MARK_FIXUP(fixup); OUT32((uint32_t)gen_helper_address[ARG(2)]);
    RESOLVE_FIXUP(PCREL, fixup);
QOP_END

QOP_START(helperf64)
    long fixup;
#ifdef HOST_USE_REGPARM
    /* movl %ebp, %eax */
    OUT8(0x8b); MODRM(HOST_REG_EAX, HOST_ENV_REG);
    /* movl $result, %edx */
    OUT8(0xb8 + HOST_REG_EDX); OUT32(arg_regno(0) - FIRST_FLOAT64_REG);
    /* movl $arg1, %ecx */
    OUT8(0xb8 + HOST_REG_ECX); OUT32(arg_regno(1) - FIRST_FLOAT64_REG);
#else
#error not implemented
#endif
    /* call helper */
    OUT8(0xe8); MARK_FIXUP(fixup); OUT32((uint32_t)gen_helper_address[ARG(2)]);
    RESOLVE_FIXUP(PCREL, fixup);
QOP_END

QOP_START(helperf64_f64)
    long fixup;
#ifdef HOST_USE_REGPARM
    /* movl %ebp, %eax */
    OUT8(0x8b); MODRM(HOST_REG_EAX, HOST_ENV_REG);
    /* movl $result, %edx */
    OUT8(0xb8 + HOST_REG_EDX); OUT32(arg_regno(0) - FIRST_FLOAT64_REG);
    /* movl $arg1, %ecx */
    OUT8(0xb8 + HOST_REG_ECX); OUT32(arg_regno(1) - FIRST_FLOAT64_REG);
    /* pushl $arg2 */
    OUT8(0x6a); OUT8(arg_regno(2) - FIRST_FLOAT64_REG);
#else
#error not implemented
#endif
    /* call helper */
    OUT8(0xe8); MARK_FIXUP(fixup); OUT32((uint32_t)gen_helper_address[ARG(3)]);
    RESOLVE_FIXUP(PCREL, fixup);
    /* Pop the argument off the stack. Faster and smaller than addl $4, %esp */
    OUT8(0x58 + HOST_REG_ECX);
QOP_END

QOP_START(helperf64_f64x32)
    long fixup;
    qemu_assert (arg_regno(0) == HOST_REG_EAX, "bad reg in helper");
#ifdef HOST_USE_REGPARM
    /* movl %ebp, %eax */
    OUT8(0x8b); MODRM(HOST_REG_EAX, HOST_ENV_REG);
    /* movl $arg1, %edx */
    OUT8(0xb8 + HOST_REG_EDX); OUT32(arg_regno(1) - FIRST_FLOAT64_REG);
    /* movl $arg2, %ecx */
    OUT8(0xb8 + HOST_REG_ECX); OUT32(arg_regno(2) - FIRST_FLOAT64_REG);
#else
#error not implemented
#endif
    /* call helper */
    OUT8(0xe8); MARK_FIXUP(fixup); OUT32((uint32_t)gen_helper_address[ARG(3)]);
    RESOLVE_FIXUP(PCREL, fixup);
QOP_END

QOP_START(start_tb)
    uint32_t val = ARG(0);
    
    if (val) {
        /* subl $val, %esp */
        OUT8(0x81);
        OUT8(0xec);
        OUT32(val);
    }
QOP_END

QOP_START(goto_tb)
    uint32_t  stack_adj;
    stack_adj = ARG(0);
#ifdef USE_DIRECT_JUMP
    /* This code must be consistent with tb_set_jmp_target.  */
    MARK_JMP(ARG(1));
    if (stack_adj) {
        /* jmp <end> */
        OUT8(0xeb); OUT8(11);
        /* addl $stack_adj, %esp */
        OUT8(0x81); OUT8(0xc4); OUT32(stack_adj);
        /* jmp <offset> */
        OUT8(0xe9); OUT32(0);
    } else {
        /* 5 bytes of nops.  */
        OUT8(0x90); OUT32(0x0026748d);
    }
#else
    long addr;
    addr = ARG(2) + offsetof(TranslationBlock, tb_next[ARG(1)]);
    
    /* This qop is a barrier, so we can clobber eax.  */
    /* movl <next_addr>, %eax */
    OUT8(0xa1); OUT32(addr);
    /* testl %eax, %eax  */
    OUT8(0x85); OUT8(0xc0);
    if (stack_adj) {
        /* jz <end> */
        OUT8(0x74); OUT8(8);
        /* addl $stack_adj, %esp */
        OUT8(0x81); OUT8(0xc4); OUT32(stack_adj);
    } else {
        /* je <end> */
        OUT8(0x74); OUT8(2);
    }
    /* jmp *%eax */
    OUT8(0xff); OUT8(0xe0);
#endif
QOP_END

QOP_START(exit_tb)
    uint32_t val = ARG(0);
    
    if (val) {
        /* addl $val, %esp */
        OUT8(0x81); OUT8(0xc4); OUT32(val);
    }
    /* ret */
    OUT8(0xc3);
QOP_END

QOP_START(st32)
    int reg;

#ifdef BSWAP_NEEDED
    reg = arg_regno(2);
    OUT8(0x8b); MODRM(reg, arg_regno(1));
    OUT8(0x0f); OUT8(0xc8 + reg);
#else
    reg = arg_regno(1);
#endif
    OUT8(0x89); OUT8(0x00 | (reg << 3) | arg_regno(0));
QOP_END

QOP_START(ld32)
    int reg;
    reg = arg_regno(0);
    OUT8(0x8b); OUT8(0x00 | (reg << 3) | arg_regno(1));
#ifdef BSWAP_NEEDED
    OUT8(0x0f); OUT8(0xc8 + reg);
#endif
QOP_END

QOP_START(ld8s32)
    int reg;
    reg = arg_regno(0);
    OUT8(0x0f); OUT8(0xbe); OUT8(0x00 | (reg << 3) | arg_regno(1));
QOP_END

QOP_START(ld8u32)
    int reg;
    reg = arg_regno(0);
    OUT8(0x0f); OUT8(0xb6); OUT8(0x00 | (reg << 3) | arg_regno(1));
QOP_END

QOP_START(st8)
    OUT8(0x88); OUT8(0x00 | (arg_regno(1) << 3) | arg_regno(0));
QOP_END

QOP_START(ld16s32)
    int reg;
    reg = arg_regno(0);
    /* movswl (%addr), %reg */
    OUT8(0x0f); OUT8(0xbf); OUT8(0x00 | (reg << 3) | arg_regno(1));
#ifdef BSWAP_NEEDED
    /* rolw $8, %regw */
    OUT8(0x66); OUT8(0xc1); MODRM(0, reg); OUT8(8);
    if (reg == HOST_REG_EAX) {
        /* cwtl */
        OUT8(0x98);
    } else {
        /* movswl %regw, %reg */
        OUT8(0x0f); OUT8(0xbf); MODRM(reg, reg);
    }
#endif
QOP_END

QOP_START(ld16u32)
    int reg;
    reg = arg_regno(0);
    /* movzwl (%addr), %reg */
    OUT8(0x0f); OUT8(0xb7); OUT8(0x00 | (reg << 3) | arg_regno(1));
#ifdef BSWAP_NEEDED
    /* rolw $8, %reg */
    OUT8(0x66); OUT8(0xc1); MODRM(0, reg); OUT8(8);
#endif
QOP_END

QOP_START(st16)
    int reg;

#ifdef BSWAP_NEEDED
    reg = arg_regno(2);
    /* mov %reg, %tmp */
    OUT8(0x8b); MODRM(reg, arg_regno(1));
    /* rolw $8, %tmp */
    OUT8(0x66); OUT8(0xc1); MODRM(0, reg); OUT8(8);
#else
    reg = arg_regno(1);
#endif
    /* movw %val, (%addr) */
    OUT8(0x66); OUT8(0x89); OUT8(0x00 | (reg << 3) | arg_regno(0));
QOP_END

QOP_START(stf64)
    int src_qreg;
    int dest_reg;
    int tmp = arg_regno(2);
    src_qreg = ARG(1);
    dest_reg = arg_regno(0);
#ifdef BSWAP_NEEDED
    OUT8(0x8b); MODRM_FLOAT64(src_qreg, tmp, 1);
    OUT8(0x0f); OUT8(0xc8 + tmp);
#else
    OUT8(0x8b); MODRM_FLOAT64(src_qreg, tmp, 0);
#endif
    OUT8(0x89); OUT8(0x00 | (tmp << 3) | dest_reg);
#ifdef BSWAP_NEEDED
    OUT8(0x8b); MODRM_FLOAT64(src_qreg, tmp, 0);
    OUT8(0x0f); OUT8(0xc8 + tmp);
#else
    OUT8(0x8b); MODRM_FLOAT64(src_qreg, tmp, 1);
#endif
    OUT8(0x89); OUT8(0x40 | (tmp << 3) | dest_reg); OUT8(4);
QOP_END

QOP_START(ldf64)
    int dest_qreg;
    int tmp = arg_regno(2);
    dest_qreg = ARG(0);
    OUT8(0x8b); OUT8(0x00 | (tmp << 3) | arg_regno(1));
#ifdef BSWAP_NEEDED
    OUT8(0x0f); OUT8(0xc8 + tmp);
    OUT8(0x89); MODRM_FLOAT64(dest_qreg, tmp, 1);
#else
    OUT8(0x89); MODRM_FLOAT64(dest_qreg, tmp, 0);
#endif
    OUT8(0x8b); OUT8(0x40 | (tmp << 3) | arg_regno(1)); OUT8(4);
#ifdef BSWAP_NEEDED
    OUT8(0x0f); OUT8(0xc8 + tmp);
    OUT8(0x89); MODRM_FLOAT64(dest_qreg, tmp, 0);
#else
    OUT8(0x89); MODRM_FLOAT64(dest_qreg, tmp, 1);
#endif
QOP_END

QOP_START(add64_split)
    /* addl src_low, dest_low */
    OUT8(0x03); MODRM(arg_regno(0), arg_regno(4));
    /* adcl src_low, dest_low */
    OUT8(0x13); MODRM(arg_regno(1), arg_regno(5));
QOP_END

QOP_START(sub64_split)
    /* subl src_low, dest_low */
    OUT8(0x2b); MODRM(arg_regno(0), arg_regno(4));
    /* sbbl src_low, dest_low */
    OUT8(0x1b); MODRM(arg_regno(1), arg_regno(5));
QOP_END

QOP_START(neg64_split)
    int reg;
    /* negl low */
    OUT8(0xf7); MODRM(3, arg_regno(0));
    reg = arg_regno(1);
    /* adcl $0, high */
    OUT8(0x83); OUT8(0xd0 | reg); OUT8(0);
    /* negl high */
    OUT8(0xf7); MODRM(3, reg);
QOP_END
QOP_CODE_END
