/*
 *  x86-64 Host code generation
 * 
 *  Copyright (c) 2005 CodeSourcery, LLC
 *  Written by Paul Brook
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* The low 3 bits of a value.  */
#define LO3(x) ((x) & 7)
/* Helper macros for handling REX prefixes.  We output a rex byte even if it
   isn't absoulutely neccessary.  */
/* DECLARE_REX is put at the start of each QOP using the following macros.  */
#define DECLARE_REX long rex_p; unsigned char rex
/* Initialize a rex byte for a 32 or 64 bit operation.  */
#define OUT_REX() do { MARK_FIXUP(rex_p); OUT8(0); rex = 0x40; } while(0)
#define OUT_REXW() do { MARK_FIXUP(rex_p); OUT8(0); rex = 0x48; } while(0)
/* Set the specified bits in the rex byte if regno > 7.  */
#define SET_REX(regno, mask) do { if ((regno) & 8) rex |= (mask); } while(0)
/* Finalize the REX byte.  */
#define FIXUP_REX() RESOLVE_FIXUP(REX, rex_p)

#define DO_FIXUP_REX(from, to) \
    *(uint8_t *)from = rex

#define HOST_REG_ENV HOST_REG_AREG0
static inline int get_arg_regno(int qreg)
{
    qemu_assert(qreg >= FIRST_HARD_QREG, "bad hard regno");
    return qreg - FIRST_HARD_QREG;
}

static inline long get_arg_im(int qreg)
{
    qemu_assert(gen_reginfo[qreg].type == QREGT_IM, "bad immediate qreg");
    return gen_reginfo[qreg].val;
}

#define arg_regno(n) get_arg_regno(ARG(n))
#define arg_im(n) get_arg_im(ARG(n))

#define DO_FIXUP_PCREL(from, to) \
    *(uint32_t *)from -= to;

/* Encode a Mod R/M byte for a register operand.  */
#define MODRM(reg, rm) do { \
    OUT8(0xc0 | (LO3(reg) << 3) | LO3(rm)); \
    SET_REX(reg, 4); \
    SET_REX(rm, 1); \
    }while(0)

QOP_CODE_START

/* Output a Mod R/M byte and offset for a base+offset memory address.  */
#define MODRM_OFFSET(reg, base, offset) do { \
    if (base == HOST_REG_RSP) {\
        if ((int8_t)(offset) == (int32_t)(offset)) { \
            OUT8(0x44 | (LO3(reg) << 3)); \
            OUT8(0x24); \
            OUT8(offset); \
        } else { \
            OUT8(0x84 | (LO3(reg) << 3)); \
            OUT8(0x24); \
            OUT32(offset); \
        } \
        SET_REX(reg, 4); \
    } else { \
        if (offset == 0 && (base) != HOST_REG_RBP) { \
            OUT8(0x00 | (LO3(reg) << 3) | LO3(base)); \
        } else if ((int8_t)(offset) == (int32_t)(offset)) { \
            OUT8(0x40 | (LO3(reg) << 3) | LO3(base)); \
            OUT8(offset); \
        } else { \
            OUT8(0x80 | (LO3(reg) << 3) | LO3(base)); \
            OUT32(offset); \
        } \
        SET_REX(reg, 4); \
        SET_REX(base, 1); \
    }}while(0);

#define DO_MOVE(bits, wide) \
QOP_START(mov##bits) \
    int dest_qreg = ARG(0); \
    int src_qreg = ARG(1); \
    struct qreg *info; \
    int regno; \
    int qreg; \
    DECLARE_REX; \
    OUT_REX##wide(); \
    qreg = QREG_NULL; \
    if (src_qreg < FIRST_HARD_QREG) { \
        info = &gen_reginfo[src_qreg]; \
        regno = arg_regno(0); \
        if (info->type == QREGT_IM) { \
            OUT8(0xb8 + LO3(regno)); OUT32(info->val); \
            if (bits == 64) \
                OUT32(0); \
            SET_REX(regno, 1); \
        } else { \
            qreg = src_qreg; \
        } \
    } else if (dest_qreg < FIRST_HARD_QREG) { \
        qreg = dest_qreg; \
        info = &gen_reginfo[dest_qreg]; \
        regno = arg_regno(1); \
    } else { \
        OUT8(0x89); MODRM(arg_regno(1), arg_regno(0)); \
    } \
    if (qreg) { \
        int val; \
        val = info->val; \
        if (qreg == src_qreg) \
            OUT8(0x8b); \
        else \
            OUT8(0x89); \
        switch (info->type) { \
        case QREGT_STACK: \
            MODRM_OFFSET(regno, HOST_REG_RSP, val); \
            break; \
        case QREGT_ENV: \
            MODRM_OFFSET(regno, HOST_REG_ENV, val); \
            break; \
        default: \
            cpu_abort(NULL, "host-i386: Bad qreg type"); \
        } \
    } \
    FIXUP_REX(); \
QOP_END

DO_MOVE(32, );
DO_MOVE(64, W);

QOP_START(mov64i)
    int reg = arg_regno(0);
    DECLARE_REX;
    OUT_REXW(); OUT8(0xb8 + LO3(reg)); SET_REX(reg, 1);
    OUT32(ARG(1)); OUT32(ARG(2));
    FIXUP_REX();
QOP_END

QOP_START(neg32)
    int reg = arg_regno(0);
    DECLARE_REX; OUT_REX();
    OUT8(0xf7); MODRM(3, reg);
    FIXUP_REX();
QOP_END

QOP_START(neg64)
    int reg = arg_regno(0);
    DECLARE_REX; OUT_REXW();
    OUT8(0xf7); MODRM(3, reg);
    FIXUP_REX();
QOP_END

QOP_START(not32)
    int reg = arg_regno(0);
    DECLARE_REX; OUT_REX();
    OUT8(0xf7); MODRM(2, reg);
    FIXUP_REX();
QOP_END

QOP_START(not64)
    int reg = arg_regno(0);
    DECLARE_REX; OUT_REXW();
    OUT8(0xf7); MODRM(2, reg);
    FIXUP_REX();
QOP_END

QOP_START(bswap32)
    int reg = arg_regno(0);
    DECLARE_REX; OUT_REX();
    OUT8(0x0f); OUT8(0xc8 + LO3(reg));
    SET_REX(reg, 1);
    FIXUP_REX();
QOP_END

QOP_START(bswap64)
    int reg = arg_regno(0);
    DECLARE_REX; OUT_REXW();
    OUT8(0x0f); OUT8(0xc8 + LO3(reg));
    SET_REX(reg, 1);
    FIXUP_REX();
QOP_END

#define DO_BINOP_WIDTH(name, op, width) \
QOP_START(name) \
    int dest = arg_regno(0); \
    int src = arg_regno(2); \
    DECLARE_REX; OUT_REX##width(); \
    OUT8(op); MODRM(dest, src); \
    FIXUP_REX(); \
QOP_END

#define DO_BINOP(name, op) \
    DO_BINOP_WIDTH(name##32, op,) \
    DO_BINOP_WIDTH(name##64, op, W)

DO_BINOP(and, 0x23)
DO_BINOP(or, 0x0b)
DO_BINOP(xor, 0x33)
DO_BINOP(add, 0x03)
DO_BINOP(sub, 0x2b)

#define DO_SHIFT_WIDTH(name, op, width) \
QOP_START(name) \
    int reg = arg_regno(0); \
    DECLARE_REX; OUT_REX##width(); \
    if (ARG(2) >= FIRST_HARD_QREG) { \
        qemu_assert(arg_regno(2) == HOST_REG_RCX, "bad shift qop"); \
        OUT8(0xd3); MODRM(op, reg); \
    } else { \
        int n; \
        n = arg_im(2); \
        if (n == 1) \
            OUT8(0xd1); \
        else \
            OUT8(0xc1); \
        MODRM(op, reg); \
        if (n != 1) \
            OUT8(n); \
    } \
    FIXUP_REX(); \
QOP_END

#define DO_SHIFT(name, op) \
    DO_SHIFT_WIDTH(name##32, op, ) \
    DO_SHIFT_WIDTH(name##64, op, W)

DO_SHIFT(shr, 5)
DO_SHIFT(sar, 7)
DO_SHIFT(shl, 4)
DO_SHIFT(ror, 1)
DO_SHIFT(rol, 0)

QOP_START(mul32)
    int dest = arg_regno(0);
    int src = arg_regno(2);
    int scratch = arg_regno(3);
    DECLARE_REX; OUT_REX();

    qemu_assert (dest == HOST_REG_RAX && scratch == HOST_REG_RDX,
                 "invalid mul qop");
    OUT8(0xf7); MODRM(4, src);
    FIXUP_REX();
QOP_END

QOP_START(mul64)
    int dest = arg_regno(0);
    int src = arg_regno(2);
    int scratch = arg_regno(3);
    DECLARE_REX; OUT_REXW();

    qemu_assert (dest == HOST_REG_RAX && scratch == HOST_REG_RDX,
                 "invalid mul qop");
    OUT8(0xf7); MODRM(4, src);
    FIXUP_REX();
QOP_END

#define DO_EXT_WIDTH(name, op, width) \
QOP_START(ext ## name) \
    int dest = arg_regno(0); \
    int src = arg_regno(1); \
    DECLARE_REX; OUT_REX##width(); \
    OUT8(0x0f); OUT8(op); MODRM(dest, src); \
    FIXUP_REX(); \
QOP_END

#define DO_EXT(name, op) \
    DO_EXT_WIDTH(name##32, op, )
//    DO_EXT_WIDTH(name##64, op, W)

DO_EXT(8u, 0xb6)
DO_EXT(16u, 0xb7)
DO_EXT(8s, 0xbe)
DO_EXT(16s, 0xbf)

QOP_START(ext32u64)
    int dest = arg_regno(0);
    int src = arg_regno(1);
    DECLARE_REX; OUT_REX();
    OUT8(0x8b); MODRM(dest, src);
    FIXUP_REX();
QOP_END

QOP_START(ext32s64)
    int dest = arg_regno(0);
    int src = arg_regno(1);
    DECLARE_REX; OUT_REXW();
    OUT8(0x63); MODRM(dest, src);
    FIXUP_REX();
QOP_END

#define DO_SETcc32(name, cc) \
QOP_START(set_##name##32) \
    int dest = arg_regno(0); \
    int src1 = arg_regno(1); \
    int src2 = arg_regno(2); \
    DECLARE_REX; \
    /* cmpl */ \
    OUT_REX(); OUT8(0x3b); MODRM(src1, src2); FIXUP_REX(); \
    /* set<cc> dest */ \
    OUT_REX(); OUT8(0x0f); OUT8(0x90 + cc); MODRM(0, dest); FIXUP_REX(); \
    /* movzbl dest, dest */ \
    OUT_REX(); OUT8(0x0f); OUT8(0xb6); MODRM(dest, dest); FIXUP_REX(); \
QOP_END

DO_SETcc32(ltu, 2)
DO_SETcc32(leu, 6)

#define DO_JMPcc32(name, op) \
QOP_START(jmp_##name##32) \
    int regno = arg_regno(0); \
    DECLARE_REX; \
    /* testl src, src */ \
    OUT_REX(); OUT8(0x85); MODRM(regno, regno); FIXUP_REX(); \
    /* j<cc> dest */ \
    OUT8(0x0f); OUT8(0x80 + op); \
    OUT32(LABEL_OFFSET(ARG(1)) - 4); \
QOP_END

DO_JMPcc32(z, 4)
DO_JMPcc32(nz, 5)
DO_JMPcc32(s, 8)
DO_JMPcc32(ns, 9)

QOP_START(jmp)
    OUT8(0xe9); OUT32(LABEL_OFFSET(ARG(0)) - 4);
QOP_END

QOP_ALIAS(helper64)
QOP_ALIAS(helper64x32)
QOP_ALIAS(helper32x64)
QOP_START(helper32)
    long fixup;
    DECLARE_REX;

    qemu_assert (arg_regno(0) == HOST_REG_RAX, "bad reg in helper");
    qemu_assert (arg_regno(1) == HOST_REG_RSI, "bad reg in helper");

    /* movq %rbp, %rdi */
    OUT_REXW(); OUT8(0x8b); MODRM(HOST_REG_RDI, HOST_REG_ENV); FIXUP_REX();

    /* call helper */
    /* ??? Need to figure out how to handle out-of-range branches.
       Probably use call stubs at the start of the translation buffer.  */
    OUT8(0xe8); MARK_FIXUP(fixup);
    OUT32((uint32_t)(long)gen_helper_address[ARG(2)]);
    RESOLVE_FIXUP(PCREL, fixup);
QOP_END

QOP_ALIAS(helper64_64)
QOP_ALIAS(helper64_64x32)
QOP_START(helper32_32)
    long fixup;
    DECLARE_REX;

    qemu_assert (arg_regno(0) == HOST_REG_RAX, "bad reg in helper");
    qemu_assert (arg_regno(1) == HOST_REG_RSI, "bad reg in helper");
    qemu_assert (arg_regno(2) == HOST_REG_RDX, "bad reg in helper");

    /* movq %rbp, %rdi */
    OUT_REXW(); OUT8(0x8b); MODRM(HOST_REG_RDI, HOST_REG_ENV); FIXUP_REX();

    /* call helper */
    /* ??? Need to figure out how to handle out-of-range branches.
       Probably use call stubs at the start of the translation buffer.  */
    OUT8(0xe8); MARK_FIXUP(fixup);
    OUT32((uint32_t)(long)gen_helper_address[ARG(3)]);
    RESOLVE_FIXUP(PCREL, fixup);
QOP_END

QOP_START(start_tb)
    uint32_t val = ARG(0);
    
    if (val) {
        /* subq $val, %rsp */
        OUT8(0x48); OUT8(0x81); OUT8(0xec); OUT32(val);
    }
QOP_END

QOP_START(goto_tb)
    uint32_t  stack_adj;
    stack_adj = ARG(0);
#ifdef USE_DIRECT_JUMP
#error needs checking
    /* This code must be consistent with tb_set_jmp_target.  */
    MARK_JMP(ARG(1));
    if (stack_adj) {
        /* jmp <end> */
        OUT8(0xeb); OUT8(12);
        /* addq $stack_adj, %rsp */
        OUT8(0x48); OUT8(0x81); OUT8(0xc4); OUT32(stack_adj);
        /* jmp <offset> */
        OUT8(0xe9); OUT32(0);
    } else {
        /* 5 bytes of nops.  */
        OUT8(0x90); OUT32(0x0026748d);
    }
#else
    long addr;
    addr = ARG(2) + offsetof(TranslationBlock, tb_next[ARG(1)]);
    
    /* This qop is a barrier, so we can clobber rax.  */
    /* movl <next_addr>, %rax */
    OUT8(0x48); OUT8(0xa1);
    OUT32((uint32_t)addr); OUT32((uint32_t)(addr >> 32));
    /* testq %rax, %rax  */
    OUT8(0x48); OUT8(0x85); OUT8(0xc0);
    if (stack_adj) {
        /* jz <end> */
        OUT8(0x74); OUT8(9);
        /* addq $stack_adj, %rsp */
        OUT8(0x48); OUT8(0x81); OUT8(0xc4); OUT32(stack_adj);
    } else {
        /* je <end> */
        OUT8(0x74); OUT8(3);
    }
    /* jmp *%rax */
    OUT8(0x48); OUT8(0xff); OUT8(0xe0);
#endif
QOP_END

QOP_START(exit_tb)
    uint32_t val = ARG(0);
    
    if (val) {
        /* addq $val, %rsp */
        OUT8(0x48); OUT8(0x81); OUT8(0xc4); OUT32(val);
    }
    /* ret */
    OUT8(0xc3);
QOP_END

/* Load/store ops */
QOP_START(st64)
    int reg;
    DECLARE_REX;
#ifdef BSWAP_NEEDED
    reg = arg_regno(2);
    OUT_REXW(); OUT8(0x8b); MODRM(reg, arg_regno(1)); FIXUP_REX();
    OUT_REXW(); OUT8(0x0f); OUT8(0xc8 + LO3(reg));
    SET_REX(reg, 1); FIXUP_REX();
#else
    reg = arg_regno(1);
#endif
    OUT_REXW(); OUT8(0x89); MODRM_OFFSET(reg, arg_regno(0), 0); FIXUP_REX();
QOP_END

QOP_START(ld64)
    int reg;
    DECLARE_REX;
    reg = arg_regno(0);
    OUT_REXW(); OUT8(0x8b); MODRM_OFFSET(reg, arg_regno(1), 0); FIXUP_REX();
#ifdef BSWAP_NEEDED
    OUT_REXW(); OUT8(0x0f); OUT8(0xc8 + LO3(reg));
    SET_REX(reg, 1); FIXUP_REX();
#endif
QOP_END

QOP_START(st32)
    int reg;
    DECLARE_REX;
#ifdef BSWAP_NEEDED
    reg = arg_regno(2);
    OUT_REX(); OUT8(0x8b); MODRM(reg, arg_regno(1)); FIXUP_REX();
    OUT_REX(); OUT8(0x0f); OUT8(0xc8 + LO3(reg)); SET_REX(reg, 1); FIXUP_REX();
#else
    reg = arg_regno(1);
#endif
    OUT_REX(); OUT8(0x89); MODRM_OFFSET(reg, arg_regno(0), 0); FIXUP_REX();
QOP_END

QOP_START(ld32)
    int reg;
    DECLARE_REX;
    reg = arg_regno(0);
    OUT_REX(); OUT8(0x8b); MODRM_OFFSET(reg, arg_regno(1), 0); FIXUP_REX();
#ifdef BSWAP_NEEDED
    OUT_REX(); OUT8(0x0f); OUT8(0xc8 + LO3(reg)); SET_REX(reg, 1); FIXUP_REX();
#endif
QOP_END

QOP_START(ld8s32)
    int reg;
    DECLARE_REX;
    reg = arg_regno(0);
    OUT_REX(); OUT8(0x0f); OUT8(0xbe); MODRM_OFFSET(reg, arg_regno(1), 0);
    FIXUP_REX();
QOP_END

QOP_START(ld8u32)
    int reg;
    DECLARE_REX;
    reg = arg_regno(0);
    OUT_REX(); OUT8(0x0f); OUT8(0xb6); MODRM_OFFSET(reg, arg_regno(1), 0);
    FIXUP_REX();
QOP_END

QOP_START(st8)
    DECLARE_REX;
    OUT_REX(); OUT8(0x88); MODRM_OFFSET(arg_regno(1), arg_regno(0), 0);
    FIXUP_REX();
QOP_END

QOP_START(ld16s32)
    int reg;
    DECLARE_REX;
    reg = arg_regno(0);
    /* movswl (%addr), %reg */
    OUT_REX(); OUT8(0x0f); OUT8(0xbf); MODRM_OFFSET(reg, arg_regno(1), 0);
    FIXUP_REX();
#ifdef BSWAP_NEEDED
    /* rolw $8, %regw */
    OUT8(0x66); OUT_REX(); OUT8(0xc1); MODRM(0, reg); OUT8(8); FIXUP_REX();
    if (reg == HOST_REG_RAX) {
        /* cwtl */
        OUT8(0x98);
    } else {
        /* movswl %regw, %reg */
        OUT_REX(); OUT8(0x0f); OUT8(0xbf); MODRM(reg, reg); FIXUP_REX();
    }
#endif
QOP_END

QOP_START(ld16u32)
    int reg;
    DECLARE_REX;
    reg = arg_regno(0);
    /* movzwl (%addr), %reg */
    OUT_REX(); OUT8(0x0f); OUT8(0xb7); MODRM_OFFSET(reg, arg_regno(1), 0);
    FIXUP_REX();
#ifdef BSWAP_NEEDED
    /* rolw $8, %reg */
    OUT8(0x66); OUT_REX(); OUT8(0xc1); MODRM(0, reg); OUT8(8); FIXUP_REX();
#endif
QOP_END

QOP_START(st16)
    int reg;
    DECLARE_REX;
#ifdef BSWAP_NEEDED
    reg = arg_regno(2);
    /* mov %reg, %tmp */
    OUT_REX(); OUT8(0x8b); MODRM(reg, arg_regno(1)); FIXUP_REX();
    /* rolw $8, %tmp */
    OUT8(0x66); OUT_REX(); OUT8(0xc1); MODRM(0, reg); OUT8(8); FIXUP_REX();
#else
    reg = arg_regno(1);
#endif
    /* movw %val, (%addr) */
    OUT8(0x66); OUT_REX(); OUT8(0x89); MODRM_OFFSET(reg, arg_regno(0), 0);
    FIXUP_REX();
QOP_END

QOP_CODE_END
