/*
 *  Host code generation
 * 
 *  Copyright (c) 2003 Fabrice Bellard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>

#include "config.h"

#define NO_CPU_IO_DEFS
#include "cpu.h"
#include "exec-all.h"
#include "disas.h"
#include "dyngen-host.h"
#include "bitmap.h"
#include "qop.h"

/* ??? This is just to keep the gen_* routines happy.  Need to figure out
   a better way.  */
static qOP *gen_throws_exception;

#include "gen-op.h"

extern int dyngen_code(uint8_t *gen_code_buf,
                       uint16_t *jmp_offsets,
                       qOP *first_qop, const long *gen_labels,
                       const struct qreg *reginfo);

qOP *gen_first_qop;
qOP *gen_last_qop;
int nb_gen_labels;
struct qreg gen_reginfo[MAX_QREGS];
int gen_last_qreg;
int gen_free_qops;

#define QOP_BUFFER_SIZE 65528
struct qop_buffer {
    struct qop_buffer *next;
    char data[QOP_BUFFER_SIZE];
};

static struct qop_buffer first_qop_buffer;
static struct qop_buffer *current_qop_buffer;
static int qop_buffer_pos;
#define MAX_GEN_LABELS 512
static long gen_labels[MAX_GEN_LABELS];

int code_copy_enabled = 1;

#ifdef DEBUG_DISAS
static const char *op_str[] = {
#define DEF(s, n, copy_size) #s,
#include "opc.h"
#ifdef HAVE_TARGET_QOPS
    NULL,
#include "qop-target.def"
#endif /* HAVE_TARGET_QOPS */
#undef DEF
};
#endif

#define MAX_QOP_ARGS 7
/* Disable optimization accross this qop.  */
/* ??? This should probably be split up into maythrow/clobber/flowcontrol.  */
#define QOPF_barrier      0x01
/* No memory operands are permitted on the qop.  */
#define QOPF_nomem        0x02
/* The qop is commitative. ie. we can safely swap the input operands.  */
#define QOPF_commutative  0x04
/* The fist operand of this qop is the stack frame size.  */
#define QOPF_unstack      0x08
/* This qop is a birary operation.  One some hosts it will be converted into
   two argument form.  */
#define QOPF_binary       0x10

/* This is to shut up compiler warnings about unused variables.  */
static inline void __use(int x)
{
}

static void
unimplemented_qop (int opcode)
{
#ifdef DEBUG_DISAS
    cpu_abort(NULL, "Unimplemented QOP: %s", op_str[opcode]);
#else
    cpu_abort(NULL, "Unimplemented QOP: %d", opcode);
#endif
}

extern void * const gen_helper_address[];

#define QOP_ALIAS(name) case INDEX_op_ ## name:
#define QOP_START(name) case INDEX_op_ ## name: {
#define QOP_CODE_START \
static inline int size_qop(int qop, uint32_t *args, \
                           const struct qreg *reginfo) \
{ \
    int qop_size = 0; \
    switch (qop) {

#define QOP_END } break;

#define QOP_CODE_END \
    case INDEX_op_label: \
    case INDEX_op_unwind: \
        qop_size = 0; \
        break; \
    default: \
        unimplemented_qop(qop); \
    } \
    return qop_size; \
}
#define ARG(n) args[n]
#define OUT8(val) do { __use(val); qop_size++; } while (0)
#define OUT32(val) do { __use(val); qop_size += 4; } while(0)
#define LABEL_OFFSET(n) 0
#define MARK_FIXUP(ptr) ptr = 0
#define RESOLVE_FIXUP(type, ptr) do {} while(0)
#define MARK_JMP(n)

#include "qop-code.h"

#undef QOP_FIRST
#undef QOP_START
#undef QOP_ALIAS
#undef QOP_END
#undef QOP_LAST
#undef ARG
#undef OUT8
#undef OUT32
#undef MARK_FIXUP
#undef RESOLVE_FIXUP
#undef MARK_JMP

#ifdef WORDS_BIGENDIAN
#define low_offset  4
#define high_offset 0
#else
#define low_offset  0
#define high_offset 4
#endif

static const struct qreg target_reginfo[TARGET_NUM_REGS] =
{
    {0, NULL, 0, QREGT_UNKNOWN, 0},
#define DEFO32(name, offset) \
    {offsetof(CPUState, offset), NULL, QMODE_I32, QREGT_ENV, QREG_NULL},
#if HOST_LONG_BITS == 64
#define DEFO64(name, offset) \
    {offsetof(CPUState, offset), NULL, QMODE_I64, QREGT_ENV, QREGT_NULL},
#else
/* We lie about the type of the qreg.  An I64 is actually implemented as a
   pair of I32 values.  */
#define DEFO64(name, offset) \
    {offsetof(CPUState, offset) + low_offset, NULL, QMODE_I32, QREGT_ENV, \
     QREGHI_##name},
#endif
#if defined(HOST_FLOAT32)
#define DEFF32(name, offset) \
    {offsetof(CPUState, offset), NULL, QMODE_F32, QREGT_ENV, QREG_NULL},
#else
#define DEFF32(name, offset) \
    {offsetof(CPUState, offset), NULL, QMODE_I32, QREGT_ENV, QREG_NULL},
#endif
#if defined(HOST_FLOAT64) || HOST_LONG_BITS < 64
#define DEFF64(name, offset) \
    {offsetof(CPUState, offset), NULL, QMODE_F64, QREGT_ENV, QREG_NULL},
#else
#define DEFF64(name, offset) \
    {offsetof(CPUState, offset), NULL, QMODE_I64, QREGT_ENV, QREG_NULL},
#endif
#define DEFR(name, reg, mode) \
    {HOST_REG_ ## reg, NULL, mode, QREGT_REG, QREG_NULL},

#include "qregs.def"
#undef DEFO32
#undef DEFO64
#undef DEFF32
#undef DEFF64
#undef DEFR
#if HOST_LONG_BITS < 64
#define DEFO32(name, offset)
#define DEFO64(name, offset) \
    {offsetof(CPUState, offset) + high_offset, NULL, QMODE_I32, QREGT_ENV, \
     QREG_NULL},
#define DEFR(name, reg, mode)
#define DEFF32(name, reg)
#define DEFF64(name, reg)
#include "qregs.def"
#undef DEFO32
#undef DEFO64
#undef DEFF32
#undef DEFF64
#undef DEFR
#endif
};
#undef low_offset
#undef high_offset

static const char * const qreg_names[TARGET_NUM_REGS] =
{
    "null", 
#define DEFO32(name, offset) #name,
#define DEFO64(name, offset) DEFO32(name, offset)
#define DEFF32(name, reg) DEFO32(name, reg)
#define DEFF64(name, reg) DEFO32(name, reg)
#define DEFR(name, reg, mode) DEFO32(name, reg)
#include "qregs.def"
#undef DEFO32
#undef DEFO64
#undef DEFF32
#undef DEFF64
#undef DEFR
#if HOST_LONG_BITS < 64
#define DEFO32(name, offset)
#define DEFO64(name, offset) #name "_HI",
#define DEFF32(name, reg)
#define DEFF64(name, reg)
#define DEFR(name, reg, mode)
#include "qregs.def"
#undef DEFO32
#undef DEFO64
#undef DEFF32
#undef DEFF64
#undef DEFR
#endif
};

#define FIRST_QOP (INDEX_op_QOP_START + 1)

static uint8_t op_nb_args[] = {
#define DEF(s, n, copy_size) n,
#include "opc.h"
#ifdef HAVE_TARGET_QOPS
    0,
#include "qop-target.def"
#endif /* HAVE_TARGET_QOPS */
#undef DEF
};

typedef struct {
    /* The qop currently being processed.  */
    qOP *qop;
    /* hard regs not used by this instruction.  */
    host_reg_bitmap reload_regs;
    /* hard regs not used as inputs.  */
    host_reg_bitmap input_free;
    /* hard regs not used as outputs.  */
    host_reg_bitmap output_free;
    /* Unused hard regs.  */
    host_reg_bitmap unallocated_regs;
    /* Allocated regs.  */
    host_reg_bitmap live_regs;
    /* The qreg assigned to each hard reg.  */
    int allocated_regs[HOST_NUM_REGS];
    /* Hard reg assigned to each oparand.  */
    int arg_regno[MAX_QOP_ARGS];
    /* 0 = operand not allocated.
       1 = Existing value used.
       2 = Reload required.  */
    int arg_reloaded[MAX_QOP_ARGS];
    /* The original qreg for input operands.  */
    int old_arg[MAX_QOP_ARGS];
} regalloc_state;

#define QIF_QREG    1
#define QIF_INPUT   2
#define QIF_OUTPUT  4
#define QIF_ALIAS   8
typedef struct {
    int nargs;
    int flags[MAX_QOP_ARGS];
    host_reg_bitmap regclass[MAX_QOP_ARGS];
    int output[MAX_QOP_ARGS];
    uint32_t *args;
    unsigned present:1;
    unsigned small_classes:1;
    unsigned barrier:1;
    unsigned unstack:1;
    unsigned move:1;
} qop_info;

typedef struct
{
  int output_mask;
  int input_mask;
} qop_def;

static qop_info dyngen_qop_info[NB_OPS - FIRST_QOP];

static qop_def qop_defs[NB_OPS - FIRST_QOP] = {
#define DEF(name, nargs, out_mask, in_mask) {out_mask, in_mask},
#include "qop.def"
#undef DEF
};

typedef struct {
    int nargs;
    uint32_t hostclass[MAX_QOP_ARGS];
} host_qop_def;

#define GRC_ARGALIAS  0x80000000
#define GRC_ARG0      (GRC_ARGALIAS + 0)
#define GRC_ARG1      (GRC_ARGALIAS + 1)
#define GRC_INT       0x40000000
static host_qop_def host_qops[NB_OPS - FIRST_QOP] = {
#define DEF0(name) \
    [INDEX_op_ ## name - FIRST_QOP] \
    = {-1, {}},
#define DEF1(name, arg0) \
    [INDEX_op_ ## name - FIRST_QOP] \
    = {1, {arg0}},
#define DEF2(name, arg0, arg1) \
    [INDEX_op_ ## name - FIRST_QOP] \
    = {2, {arg0, arg1}},
#define DEF3(name, arg0, arg1, arg2) \
    [INDEX_op_ ## name - FIRST_QOP] \
    = {3, {arg0, arg1, arg2}},
#define DEF4(name, arg0, arg1, arg2, arg3) \
    [INDEX_op_ ## name - FIRST_QOP] \
    = {4, {arg0, arg1, arg2, arg3}},
#define DEF5(name, arg0, arg1, arg2, arg3, arg4) \
    [INDEX_op_ ## name - FIRST_QOP] \
    = {5, {arg0, arg1, arg2, arg3, arg4}},
#define DEF6(name, arg0, arg1, arg2, arg3, arg4, arg5) \
    [INDEX_op_ ## name - FIRST_QOP] \
    = {6, {arg0, arg1, arg2, arg3, arg4, arg5}},
#define DEF7(name, arg0, arg1, arg2, arg3, arg4, arg5, arg6) \
    [INDEX_op_ ## name - FIRST_QOP] \
    = {7, {arg0, arg1, arg2, arg3, arg4, arg5, arg6}},
#include "qop-host.def"
#undef DEF7
#undef DEF6
#undef DEF5
#undef DEF4
#undef DEF3
#undef DEF2
#undef DEF1
};

/* Obtain register constraint information about a qop.  */
static inline qop_info *match_qop(qOP *qop)
{
    qop_info *qi;
    qemu_assert(qop->opcode >= FIRST_QOP, "Tried to match legacy op");
    qi = &dyngen_qop_info[qop->opcode - FIRST_QOP];
    qi->args = qop->args;
    return qi;
}

static const int qop_unstacks[] = {
    INDEX_op_start_tb,
    INDEX_op_exit_tb,
    INDEX_op_goto_tb,
    0
};

static const int qop_barriers[] = {
    INDEX_op_helper32,
    INDEX_op_helper32_32,
    INDEX_op_helper64,
    INDEX_op_helper64x32,
    INDEX_op_helper32x64,
    INDEX_op_helper64_64,
    INDEX_op_helper64_64x32,
    INDEX_op_helperf64,
    INDEX_op_helperf64x32,
    INDEX_op_helper32xf64,
    INDEX_op_helperf64_f64,
    INDEX_op_helperf64_f64x32,
    INDEX_op_jmp_z32,
    INDEX_op_jmp_nz32,
    INDEX_op_jmp_s32,
    INDEX_op_jmp_ns32,
    INDEX_op_jmp,
    INDEX_op_start_tb,
    INDEX_op_exit_tb,
    INDEX_op_goto_tb,
    INDEX_op_label,
    0
};

static const int qop_moves[] = {
    INDEX_op_mov32,
    INDEX_op_mov64,
    INDEX_op_mov64i,
    INDEX_op_movf32,
    INDEX_op_movf64,
    0
};

/* Initialize dyngen_qop_info for match_qop.  */
static void dyngen_init_qop_info(void)
{
    int n;
    int i;
    qop_info *qi;
    qop_def *def;
    host_qop_def *hdef;
    int old_nargs;
    int arg;
    uint32_t hostclass;

    init_host_reg_classes();
    for (n = 0; n < NB_OPS - FIRST_QOP; n++) {
        qi = &dyngen_qop_info[n];
        def = &qop_defs[n];
        hdef = &host_qops[n];
        old_nargs = op_nb_args[FIRST_QOP + n];
        if (hdef->nargs == 0) {
            /* Include enough information for debug dumps to work.  */
            qi->nargs = old_nargs;
            for (arg = 0; arg < qi->nargs; arg++) {
                if ((def->input_mask | def->output_mask) & (1 << arg))
                    qi->flags[arg] = QIF_QREG;
            }
            continue;
        }

        if (hdef->nargs == -1)
            i = 0;
        else
            i = hdef->nargs;
        op_nb_args[FIRST_QOP + n] = i;
        qi->nargs = i;
        qemu_assert(i >= old_nargs, "host qop does not have enough args");
        qemu_assert(!qi->present, "duplicate qop definition");
        qi->present = 1;
        qi->small_classes = 0;
        /* An argument must not be both an input and an oputput.  */
        qemu_assert((def->input_mask & def->output_mask) == 0,
                    "inout qop argument");
        for (arg = 0; arg < hdef->nargs; arg++) {
            hostclass = hdef->hostclass[arg];
            if (arg < old_nargs) {
                /* Normal argument.  */
                if (def->input_mask & (1 << arg)) {
                    /* Aliases are marked on inputs because that makes most
                       sense to me when writing the host descriptions, but
                       the code expects the active argument to be the input.
                       Move stuff around to make this happen.  */
                    if (hostclass & GRC_ARGALIAS) {
                        int alias;
                        alias = hostclass & ~GRC_ARGALIAS;
                        qi->output[arg] = alias;
                        qi->flags[alias] = QIF_ALIAS;
                        qi->flags[arg] = QIF_QREG | QIF_INPUT | QIF_OUTPUT;
                        qi->regclass[arg] = qi->regclass[alias];
                    } else {
                        qi->flags[arg] = QIF_QREG | QIF_INPUT;
                    }
                }
                if (def->output_mask & (1 << arg)) {
                    qi->flags[arg] = QIF_QREG | QIF_OUTPUT;
                }
            } else {
                /* Scratch register.  */
                qi->flags[arg] = QIF_QREG | QIF_INPUT | QIF_OUTPUT;
            }
            if ((hostclass & GRC_ARGALIAS) == 0) {
                HBM_CLEAR(qi->regclass[arg]);
                for (i = 0; i < NUM_HOST_REG_CLASSES; i++) {
                    if (hostclass & (1 << i)) {
                        HBM_OR(qi->regclass[arg], qi->regclass[arg],
                               host_reg_classes[i]);
                    }
                }
                if (HBM_IS_SINGLE_REG(qi->regclass[arg]))
                    qi->small_classes = 1;
            }
        }
    }
    /* Mark qops with side-effects.  */
    for (i = 0; qop_unstacks[i]; i++)
        dyngen_qop_info[qop_unstacks[i] - FIRST_QOP].unstack = 1;
    for (i = 0; qop_barriers[i]; i++)
        dyngen_qop_info[qop_barriers[i] - FIRST_QOP].barrier = 1;
    for (i = 0; qop_moves[i]; i++)
        dyngen_qop_info[qop_moves[i] - FIRST_QOP].move = 1;
}

#ifdef DEBUG_DISAS
static const unsigned short opc_copy_size[] = {
#define DEF(s, n, copy_size) copy_size,
#include "opc.h"
#undef DEF
};


#define DEBUG_BUFFER_SIZE 1024000
char debug_buffer[DEBUG_BUFFER_SIZE];
int debug_buffer_pos = 0;

void dump_ops(qOP *first_qop)
{
    int c, n, i;
    qOP *qop;
    int opnum;
    qop_info *qi;
    int reg;

    opnum = 0;
    for (qop = first_qop; qop; qop = qop->next) {
        c = qop->opcode;
        if (c >= FIRST_QOP && c < NB_OPS)
            qi = match_qop(qop);
        else
            qi = NULL;
        n = op_nb_args[c];
        fprintf(logfile, "0x%04x: %s", opnum++,  op_str[c]);
        for(i = 0; i < n; i++) {
            if (qi && (qi->flags[i] & (QIF_QREG | QIF_ALIAS)) != 0) {
                reg = qop->args[i];
                if (reg < 0
                    || (reg >= gen_last_qreg && reg < FIRST_HARD_QREG)
                    || (reg >= FIRST_HARD_QREG + HOST_NUM_REGS)) {
                    fprintf (logfile, " Invalid(0x%x)", reg);
                } else if (reg < TARGET_NUM_REGS) {
                    fprintf(logfile, " %s", qreg_names[reg]);
                } else if (reg < FIRST_HARD_QREG) {
                    switch (gen_reginfo[reg].type) {
                    case QREGT_UNKNOWN:
                        fprintf(logfile, " tmp%d",
                                qop->args[i] - TARGET_NUM_REGS);
                        break;
                    case QREGT_REG:
                        fprintf(logfile, " reg%d",
                                (int)gen_reginfo[reg].val);
                        break;
                    case QREGT_STACK:
                        fprintf(logfile, " stack(%d)",
                                (int)gen_reginfo[reg].val);
                        break;
                    case QREGT_IM:
                        fprintf(logfile, " 0x%x",
                                (int)gen_reginfo[reg].val);
                        break;
                    default:
                        fprintf(logfile, " ???");
                        break;
                    }
                } else {
                    fprintf(logfile, " hreg%d", reg - FIRST_HARD_QREG);
                }
            } else {
                fprintf(logfile, " 0x%x", qop->args[i]);
            }
        }
        fprintf(logfile, "\n");
    }
}

#endif

/* Create a new qop.  */
qOP *allocate_qop(int nargs, uint16_t opcode)
{
    qOP *new_qop;
    int size;

    if (gen_free_qops)
        gen_free_qops--;
    /* ??? We deliberately ignore the passed nargs for qops.
       Should probably remove the function argument.  */
    nargs = op_nb_args[opcode];
    size = sizeof(qOP) + nargs * sizeof(uint32_t);
    if (qop_buffer_pos + size > QOP_BUFFER_SIZE) {
        if (!current_qop_buffer->next) {
            current_qop_buffer->next = malloc(sizeof(struct qop_buffer));
            current_qop_buffer->next->next = NULL;
        }
        current_qop_buffer = current_qop_buffer->next;
        qop_buffer_pos = 0;
    }
    new_qop = (qOP *)&current_qop_buffer->data[qop_buffer_pos];
    qop_buffer_pos += size;
    memset(new_qop, 0, size);
    
    new_qop->opcode = opcode;

    return new_qop;
}

/* Determine qreg livefimes.  */
static void dyngen_reglife(void)
{
    int i;
    qOP *qop;
    int reg;
    uint16_t opc;
    qop_info *qi;

    /* ??? This assumes linear control flow.  ie. forward jumps are ok, but
       backwards jumps will break horribly.  Need to implement some sort
       of subdivision or control flow graph, or something.  */
    /* Determine register lifetimes.  */
    for (qop = gen_first_qop; qop; qop = qop->next) {
        opc = qop->opcode;
        if (opc < FIRST_QOP)
            continue;

        qi = match_qop(qop);
        for (i = 0; i < qi->nargs; i++) {
            if (qi->flags[i] & QIF_QREG) {
                gen_reginfo[qi->args[i]].last_seen = qop;
            }
        }
    }
    /* Mark register values dead on their last use.  */
    for (reg = TARGET_NUM_REGS; reg < gen_last_qreg; reg++) {
        qop = gen_reginfo[reg].last_seen;
        if (!qop)
            continue;

        qi = match_qop(qop);
        for (i = 0; i < qi->nargs; i++) {
            if (qi->flags[i] & QIF_QREG
                && qi->args[i] == reg) {
                qop->dead_regs |= 1 << i;
            }
        }
    }
}

/* Insert a qreg copy before the given qop.  */
static void insert_mov_before(qOP *qop, int to, int from, qreg_mode mode)
{
    qOP *new_qop;
    int opcode;

    switch (mode) {
    case QMODE_I32:
        opcode = INDEX_op_mov32;
        break;
    case QMODE_I64:
        opcode = INDEX_op_mov64;
        break;
    case QMODE_F32:
        opcode = INDEX_op_movf32;
        break;
    case QMODE_F64:
        opcode = INDEX_op_movf64;
        break;
    default:
        cpu_abort(NULL, "Bad mode for mov");
    }

    new_qop = allocate_qop(2, opcode);
    new_qop->args[0] = to;
    new_qop->args[1] = from;
    new_qop->prev = qop->prev;
    new_qop->next = qop;
    if (qop->prev)
        qop->prev->next = new_qop;
    qop->prev = new_qop;
}

/* Return nonzero if arg I is an inout argument with in!=out, and the input
   is still live after the assignment.  */
static int mismatched_inout(regalloc_state *rs, qop_info *qi, int i)
{
  int arg;

  arg = qi->args[i];
  if (!arg)
      return 0;
  if ((qi->flags[i] & (QIF_INPUT | QIF_OUTPUT)) != (QIF_INPUT | QIF_OUTPUT))
      return 0;
  if (arg == qi->args[qi->output[i]])
      return 0;
  if (rs->qop->dead_regs & (1 << i))
      return 0;
  return 1;
}
/* Allocate a hard reg to an argument.  */
static void allocate_reg(regalloc_state *rs, qop_info *qi, int i)
{
    int regno = -1;
    struct qreg *info;
    host_reg_bitmap tmp;

    info = &gen_reginfo[qi->args[i]];
    qemu_assert (!rs->arg_reloaded[i], "argument already allocated");
    if (info->is_live) {
        int avail;

        /* Use the existing live value directly if possible.  */
        regno = gen_reginfo[qi->args[i]].live_reg;
        /* ??? This should also test if this is a fixed reg
           (ie. dyngen style global register variable).  */
        if (HBM_TEST_REG(qi->regclass[i], regno)) {
            avail = 1;
            if ((qi->flags[i] & QIF_INPUT) != 0
                && !HBM_TEST_REG(rs->input_free, regno))
                avail = 0;
            if ((qi->flags[i] & QIF_OUTPUT) != 0
                && !HBM_TEST_REG(rs->output_free, regno))
                avail = 0;
            /* Force a reload if we this is an inout operand, the input value
               is still live after this opeartaion, and we have a convenient
               reg available.  */
            HBM_AND(tmp, rs->unallocated_regs, qi->regclass[i]);
            if (!HBM_IS_EMPTY(tmp) && mismatched_inout(rs, qi, i))
                avail = 0;
            if (avail)
                rs->arg_reloaded[i] = 1;
        }
    } else if (info->type == QREGT_REG) {
        /* The value is a fixed register. Try to use that.  */
        regno = info->val;
        /* Only allow inout fixed regs if the real output reg is also the
           same fixed reg.  */
        if (qi->args[i]
            && HBM_TEST_REG(qi->regclass[i], regno)
            && !mismatched_inout(rs, qi, i)) {
            rs->arg_reloaded[i] = 1;
        }
    }
    if (!rs->arg_reloaded[i]) {
        /* The existing value is either already in use, or
           not suitable.  Allocate a new hard reg for it.  */

        /* Try to allocate free regs first.  */
        HBM_AND(tmp, rs->unallocated_regs, qi->regclass[i]);
        if (HBM_IS_EMPTY(tmp)) {
            HBM_COPY(tmp, qi->regclass[i]);
            if ((qi->flags[i] & QIF_INPUT))
                HBM_AND(tmp, tmp, rs->input_free);
            if ((qi->flags[i] & QIF_OUTPUT))
                HBM_AND(tmp, tmp, rs->output_free);
            qemu_assert (!HBM_IS_EMPTY(tmp), "Register allocation failed");
        }
        regno = HBM_FIRST_REGNO(tmp);
        rs->arg_reloaded[i] = 2;
    }
    rs->arg_regno[i] = regno;
    if ((qi->flags[i] & QIF_INPUT))
      HBM_CLEAR_REG(rs->input_free, regno);
    if ((qi->flags[i] & QIF_OUTPUT))
        HBM_CLEAR_REG(rs->output_free, regno);
    HBM_CLEAR_REG(rs->unallocated_regs, regno);
    HBM_CLEAR_REG(rs->reload_regs, regno);
}

/* Record that a live qreg died.  */
static inline void do_reg_die(regalloc_state *rs, int qreg)
{
    struct qreg *info;
    int regno;

    info = &gen_reginfo[qreg];
    qemu_assert(info->is_live, "killed dead reg");
    regno = info->live_reg;
    HBM_CLEAR_REG(rs->live_regs, regno);
    info->is_live = 0;
}

/* Get a qreg for a hard register.  */
static inline int get_hard_qreg(int regno)
{
    return FIRST_HARD_QREG + regno;
}

/* Spill a live value back to its memory location.  */
static void do_spill(regalloc_state *rs, int regno)
{
    int qreg;
    struct qreg *info;

    qreg = rs->allocated_regs[regno];
    info = &gen_reginfo[qreg];
    qemu_assert(info->is_live, "spilled dead reg");
    qemu_assert(info->live_reg == regno, "bad live register info");
    do_reg_die(rs, qreg);
    if (info->type == QREGT_UNKNOWN)
        info->type = QREGT_STACK;
    else if (info->type == QREGT_REG)
        qreg = get_hard_qreg(info->val);
    insert_mov_before(rs->qop, qreg, get_hard_qreg(regno), info->mode);
}


/* Spill all live values back to their memory locations.  */
static void spill_everything(regalloc_state *rs)
{
    /* Force all live values back to memory.  */
    while (!HBM_IS_EMPTY(rs->live_regs)) {
        do_spill(rs, HBM_FIRST_REGNO(rs->live_regs));
    }
}

/* Associate the live value REGNO with QREG.  */
static void set_live(regalloc_state *rs, int qreg, int regno)
{
    struct qreg *info;

    /* Some outputs may be ignored.  */
    qemu_assert(qreg != QREG_NULL, "live discarded value");
    qemu_assert(!HBM_TEST_REG(rs->live_regs, regno), "reg already live");
    info = &gen_reginfo[qreg];
    if (info->is_live)
        do_reg_die(rs, qreg);
    /* Set the assigned register as the live value for this reg.  */
    if (HBM_TEST_REG(allocatable_regs, regno)) {
        info->is_live = 1;
        info->live_reg = regno;
        HBM_SET_REG(rs->live_regs, regno);
        rs->allocated_regs[info->live_reg] = qreg;
    }
}

/* Remove the current qop.  */
static inline void nullify_qop(regalloc_state *rs)
{
    /* ??? Should we remove it properly?  */
    rs->qop->opcode = INDEX_op_nop;
}

/* Pick a register out of mask to spill.  We don't have enough information
   to be clever, so just try and rotate the spilled registers.  */
static host_reg_bitmap spill_lru_mask;

static int pick_spill_reg(host_reg_bitmap mask)
{
  host_reg_bitmap tmp;
  int regno;

  HBM_AND(tmp, spill_lru_mask, mask);
  if (HBM_IS_EMPTY(tmp)) {
      HBM_OR(spill_lru_mask, spill_lru_mask, mask);
      HBM_COPY(tmp, spill_lru_mask);
  }
  regno = HBM_FIRST_REGNO(tmp);
  HBM_CLEAR_REG(spill_lru_mask, regno);
  return regno;
}

/* Perform register allocation for a simple move.  */
static void do_regalloc_mov(regalloc_state *rs, qop_info *qi)
{
    struct qreg *info0;
    struct qreg *info1;
    int regno;
    host_reg_bitmap tmp;

    info0 = &gen_reginfo[qi->args[0]];
    info1 = &gen_reginfo[qi->args[1]];
    /* This is required for correctness.  The code below assumes the source
       and destination are distinct.  Ideally these would be stripped
       earlier.  */
    if (info0 == info1) {
        nullify_qop(rs);
        return;
    }
    /* Mark the existing dest value as dead.  */
    if (info0->is_live)
        do_reg_die(rs, qi->args[0]);

    if (info1->is_live) {
        int arg;

        arg = qi->args[0];
        /* The source register is a live value.  */
        if (rs->qop->dead_regs & 2) {
            /* Origial reg is dead.   Reuse it.  */
            regno = info1->live_reg;
            do_reg_die(rs, qi->args[1]);
            nullify_qop(rs);
            set_live(rs, arg, regno);
        } else if (info0->type == QREGT_REG) {
            /* Set fixed register.  */
            qi->args[1] = get_hard_qreg(info1->live_reg);
            qi->args[0] = get_hard_qreg(info0->val);
        } else {
            HBM_BIC(tmp, allocatable_regs, rs->live_regs);
            HBM_AND(tmp, tmp, qi->regclass[1]);
            if (HBM_IS_EMPTY(tmp)) {
                /*  No free regs.  Spill the original value.  */
                regno = info1->live_reg;
                do_spill(rs, regno);
                nullify_qop(rs);
                set_live(rs, arg, regno);
            } else {
                /* Allocate new variable and set it.  */
                regno = HBM_FIRST_REGNO(tmp);
                qi->args[1] = get_hard_qreg(info1->live_reg);
                qi->args[0] = get_hard_qreg(regno);
                set_live(rs, arg, regno);
            }
        }
    } else {
        /* Source value is not live.  */
        if (info0->type == QREGT_REG) {
            /* Use fixed registers directly.  */
            regno = info0->val;
        } else {
            /* Allocate a new value and copy.  */
            HBM_BIC(tmp, allocatable_regs, rs->live_regs);
            HBM_AND(tmp, tmp, qi->regclass[0]);
            if (HBM_IS_EMPTY(tmp)) {
                /* No free regs.  Spill something.  */
                /* ??? Be cleverer about picking something to spill.  */
                /* ??? Make sure the register mode is correct.  */
                HBM_AND(tmp, rs->live_regs, qi->regclass[0]);
                regno = pick_spill_reg(tmp);
                do_spill(rs, regno);
            } else {
                /* Allocate new hard reg.  */
                regno = HBM_FIRST_REGNO(tmp);
            }
        }
        /* Set the value.  */
        set_live(rs, qi->args[0], regno);
        qi->args[0] = get_hard_qreg(regno);
        /* If the source is a fixed register replace it by the equivalent
           hard reg.  */
        if (info1->type == QREGT_REG) {
            qi->args[1] = get_hard_qreg(info1->val);
        }
    }
}

/* Assign hard regs.  */
static void dyngen_regalloc(void)
{
    regalloc_state rs;
    uint16_t opcode;
    qOP *qop;
    qop_info *qi;
    int i;
    int arg;

    HBM_CLEAR(rs.live_regs);
    for (qop = gen_first_qop; qop; qop = qop->next) {
        opcode = qop->opcode;

        /* Legacy dygen ops clobber everything.  */
        rs.qop = qop;
        if (opcode < FIRST_QOP) {
            spill_everything(&rs);
            continue;
        }

        qi = match_qop(qop);
        if (!qi->present)
            cpu_abort(NULL, "unimplemented qop (%d %s)", qop->opcode,
#ifdef DEBUG_DISAS
                      op_str[qop->opcode]);
#else
                      "");
#endif
        rs.qop = qop;
        /* Special case moves.  */
        /* ??? mov64i is wrong.  */
        if (qi->move) {
            do_regalloc_mov(&rs, qi);
            continue;
        }
        
        /* We assume we only have two classes of operand restrictions:
           - Single register classes.
           - Large classes, where "large" means we don't have to worry about
             allocation order within the instruction.

             The allocation algorithm is rather dumb, so care must be taken
             when writing host description to ensure this assumption holds.
             
             For example, a machine has 3 registers.  Class A includes all
             3 regs, class B includes only regs 1 and 2.  This can fail if 
             the following three-input instruction is used:
               DEF3(frob_three_inputs, QRC_A, QRC_A, QRC_B)
             The allocator may allocate regs 1 and 2 to the first 2
             operands, then be unable to satisfy the last constraint.  */

        HBM_COPY(rs.reload_regs, allocatable_regs);
        HBM_COPY(rs.input_free, allocatable_regs);
        HBM_COPY(rs.output_free, allocatable_regs);
        for (i = 0; i < qi->nargs; i++) {
            rs.arg_regno[i] = -1;
            rs.arg_reloaded[i] = 0;
        }
        if (qi->small_classes) {
            /* Allocate single register constraints.  */
            for (i = 0; i < qi->nargs; i++) {
                if (!(qi->flags[i] & QIF_QREG))
                    continue;
                if (!HBM_IS_SINGLE_REG(qi->regclass[i]))
                    continue;
                rs.arg_regno[i] = HBM_FIRST_REGNO(qi->regclass[i]);
                rs.arg_reloaded[i] = 2;
                HBM_CLEAR_REG(rs.reload_regs, rs.arg_regno[i]);
                if (qi->flags[i] & QIF_INPUT)
                    HBM_CLEAR_REG(rs.input_free, rs.arg_regno[i]);
                if (qi->flags[i] & QIF_OUTPUT)
                    HBM_CLEAR_REG(rs.output_free, rs.arg_regno[i]);
            }
        }
        HBM_BIC(rs.unallocated_regs, rs.reload_regs, rs.live_regs);
        /* Allocate input registers.  */
        for (i = 0; i < qi->nargs; i++) {
            if (!(qi->flags[i] & QIF_QREG))
                continue;
            if (!(qi->flags[i] & QIF_INPUT))
                continue;
            if (rs.arg_reloaded[i])
                continue;
            allocate_reg(&rs, qi, i);
        }
        /* Allocate output registers.  */
        for (i = 0; i < qi->nargs; i++) {
            if (!(qi->flags[i] & QIF_QREG))
                continue;
            if (rs.arg_reloaded[i])
                continue;
            allocate_reg(&rs, qi, i);
        }
        /* Perform input reloads.  */
        for (i = 0; i < qi->nargs; i++) {
            struct qreg *info;
            int hardreg;
            if (!(qi->flags[i] & QIF_QREG))
                continue;
            if (!(qi->flags[i] & QIF_INPUT))
                continue;
            arg = qi->args[i];
            rs.old_arg[i] = arg;
            hardreg = get_hard_qreg(rs.arg_regno[i]);

            info = &gen_reginfo[arg];
            /* The existing value was ok.  However it may have been
               spilled by an earlier input reload.  */
            if (!info->is_live) {
                if (info->type != QREGT_REG
                    || info->val != rs.arg_regno[i]) {
                    /* Value was spilled.  Reload it.  */
                    rs.arg_reloaded[i] = 2;
                }
            }
            if (rs.arg_reloaded[i] == 2) {
                /* If the register allocated is live then spill that value
                   first.  */
                /* ??? If inputs overlap this will create unneccessary
                   spills.  */
                if (HBM_TEST_REG(rs.live_regs, rs.arg_regno[i])) {
                   do_spill(&rs, rs.arg_regno[i]);
                }
                if (info->is_live)
                    arg = get_hard_qreg(info->live_reg);
                else if (info->type == QREGT_REG)
                    arg = get_hard_qreg(info->val);
                /* Copy the value to the allocated register.  */
                if (arg)
                    insert_mov_before(qop, hardreg, arg, info->mode);
            }
            qi->args[i] = hardreg;
        }
        /* Update input liveness information.  */
        for (i = 0; i < qi->nargs; i++) {
            if ((qop->dead_regs & (1 << i)) == 0)
                continue;
            arg = rs.old_arg[i];
            if (arg && gen_reginfo[arg].is_live)
                do_reg_die(&rs, arg);
        }
        /* ??? This assumes worst-case, ie. everything must be forced back
           to memory.  In many cases we know that eg. just volatile
           registers are clobbered, or only target CPUState needs to be
           forced to memory.  */
        if (qi->barrier) {
            spill_everything(&rs);
        }
        /* Perform output reloades, and associate outputs with their
           values.  */
        for (i = 0; i < qi->nargs; i++) {
            if (!(qi->flags[i] & QIF_QREG))
                continue;
            if (!(qi->flags[i] & QIF_OUTPUT))
                continue;
            if (qi->flags[i] & QIF_INPUT) {
                /* Skip scratch values.  */
                if (rs.old_arg[i] == 0)
                    continue;
                /* An inout operand is actually two operands, but the host
                   requires that they be the same in this insn.  */
                arg = qi->args[qi->output[i]];
                qi->args[qi->output[i]] = qi->args[i];
            } else {
                arg = qi->args[i];
                qi->args[i] = get_hard_qreg(rs.arg_regno[i]);
            }
            /* Mark the old value as dead.  */
            if (gen_reginfo[arg].is_live)
                do_reg_die(&rs, arg);
            /* If the register is live then spill it before it is
               overwritten.  */
            if (HBM_TEST_REG(rs.live_regs, rs.arg_regno[i])) {
               do_spill(&rs, rs.arg_regno[i]);
            }
            if (arg)
              set_live(&rs, arg, rs.arg_regno[i]);
        }
    }

    while (gen_first_qop->prev)
        gen_first_qop = gen_first_qop->prev;
}

/* compute label info */
static void dyngen_labels(long *gen_labels, int nb_gen_labels,
                          uint8_t *gen_code_buf, qOP *first_qop)
{
    uint8_t *gen_code_ptr;
    int c;
    qOP *qop;

    if (nb_gen_labels == 0)
        return;

    if (nb_gen_labels > MAX_GEN_LABELS)
        cpu_abort(NULL, "too may labels");
    /* compute the address of each op code */
    gen_code_ptr = gen_code_buf;
    for (qop = first_qop; qop; qop = qop->next) {
        c = qop->opcode;
        if (c == INDEX_op_label) {
            gen_labels[qop->args[0]] = (long)gen_code_ptr;
        } else if (c >= FIRST_QOP) {
            gen_code_ptr += size_qop(c,  qop->args, gen_reginfo);
        } else {
            gen_code_ptr += opc_copy_size[c];
        }
    }
}

/* Assign stack slots, and make sure the stack is adjusted properly.  */
static void dyngen_alloc_stack(TranslationBlock *tb)
{
    qOP *qop;
    int i;
    uint32_t stack_size;
    uint32_t size;
    int c;
    qop_info *qi;
    struct qreg *info;

    stack_size = 0;
    for (i = TARGET_NUM_REGS; i < gen_last_qreg; i++) {
        info = &gen_reginfo[i];
        qemu_assert(!info->is_live, "undead register");
        switch (info->type) {
        case QREGT_UNKNOWN:
            /* Unused qregs can occur when lowering 64-bit code.  Ideally
               we would eliminate these redundant operations.  Give them a
               value to stop the code generator barfing.  */
            info->type = QREGT_IM;
#ifdef DEBUG_DISAS
            info->val = 0xdeadbeef;
#else
            info->val = 0;
#endif
            break;
        case QREGT_STACK:
            /* ??? The stack slot allocation algoritm is really dumb.  */
            switch (gen_reginfo[i].mode) {
            case QMODE_I32:
            case QMODE_F32:
                size = 4;
                break;
            case QMODE_I64:
            case QMODE_F64:
                size = 8;
                break;
            default:
                qemu_assert(0, "Bad mode for stack qop");
            }
            /* Make sure the stack slot is correctly aligned.  */
            stack_size = (stack_size + size - 1) & ~(size - 1);
            gen_reginfo[i].val = stack_size;
            stack_size += size;
            break;
        default: /* do nothing */
            break;
        }
    }
    if (stack_size) {
        for (qop = gen_first_qop; qop; qop = qop->next) {
            c = qop->opcode;
            if (c < FIRST_QOP)
                continue;
            qi = match_qop(qop);
            if (qi->unstack)
                qop->args[0] = stack_size;
        }
    }
    qop = allocate_qop(1, INDEX_op_start_tb);
    qop->args[0] = stack_size;
    qop->next = gen_first_qop;
    gen_first_qop->prev = qop;
    gen_first_qop = qop;
    if (stack_size)
        tb->cflags |= CF_STACK_USED;
}

static inline int reg_is_mem(int reg)
{
    int type;
    type = gen_reginfo[reg].type;
    return (type == QREGT_ENV || type == QREGT_STACK);
}

#if 0
/* Remove a qop from the chain.  Does not reset the next and prev fields of
   the removed qop.  */
static void remove_qop(qOP *qop)
{
    if (qop->prev)
        qop->prev->next = qop->next;
    else
        gen_first_qop = qop->next;
    if (qop->next)
        qop->next->prev = qop->prev;
}

/* Change a qop into a simple mov const.  */
static void simplify_op(qOP *qop, uint32_t newval)
{
    qop->args[1] = gen_im32(newval);
    qop->dead_regs = 2;
    qop->opcode = INDEX_op_mov32;
}

/* This is currently broken, and probably mostly redundant now we have a
   better register allocator.  */
/* Copy and constant propagation.  */
static void dyngen_cprop(void)
{
    qOP *qop;
    qOP *next_op;
    qOP *last_op;
    int nregs;
    int c;
    int i;
    uint32_t flags;
    int src_dead;
    int dest_dead;
    int aggressive;
    int srcreg;
    int destreg;
    uint32_t newval;
    int is_const;

    for (qop = gen_first_qop; qop; qop = qop->next) {
        c = qop->opcode;
        if (c < FIRST_QOP)
            continue;
        qi = match_qop(qi);
        if (qi->nargs < 2)
            continue;

        if ((nregs > 1 && gen_reginfo[qop->args[1]].type == QREGT_IM)
            && (nregs == 2 || gen_reginfo[qop->args[2]].type == QREGT_IM))
            is_const = 1;
        else
            is_const = 0;

        if (is_const) {
            switch (c) {
            case INDEX_op_add32:
                newval = gen_reginfo[qop->args[1]].val
                         + gen_reginfo[qop->args[2]].val;
                simplify_op(qop, newval);
                break;
            case INDEX_op_sub32:
                newval = gen_reginfo[qop->args[1]].val
                         - gen_reginfo[qop->args[2]].val;
                simplify_op(qop, newval);
                break;
            case INDEX_op_neg32:
                simplify_op(qop, -gen_reginfo[qop->args[1]].val);
                break;
            case INDEX_op_not32:
                simplify_op(qop, ~gen_reginfo[qop->args[1]].val);
                break;
            }
        }
        c = qop->opcode;
        if (c != INDEX_op_mov32) {
            continue;
        }
        destreg = qop->args[0];
        srcreg = qop->args[1];

        /* Remove noop moves.  */
        if (destreg == srcreg) {
            remove_qop(qop);
            continue;
        }

        /* ??? It's probably worthwhile to split memory-memory moves early
           so that the temporary can be propagated into later uses.  */
        /* Don't bother propagating a store from memory to a temporary.  */
        if (reg_is_mem(srcreg))
            continue;

        /* Constant values and stores to memory are worth aggressively
           propagating.  */
        if (reg_is_mem(destreg) || gen_reginfo[srcreg].type == QREGT_IM) {
            aggressive = 1;
        } else {
            aggressive = 0;
        }

        dest_dead = 0;
        src_dead = (qop->dead_regs & 2) != 0;
        last_op = qop;
        qop->dead_regs = 0;
        for (next_op = qop->next; next_op; next_op = next_op->next) {
            if (next_op->opcode < FIRST_QOP)
                break;;
            c = next_op->opcode - FIRST_QOP;
            /* ??? We should be able to propagate past some types of barriers.
               For now it's not worth the effort, it'll just confuse the
               register allocator.  */
            flags = qop_flags[c];
            if (flags & QOPF_barrier)
                break;

            nregs = qop_nregs[c];
            if (nregs == 0)
                continue;

            if (next_op->args[0] == destreg) {
                /* The insn overrites the value.  */
                /* ??? If this is a guest reg we can turn it into a
                   teporary.  */
                dest_dead = 1;
            }
            for (i = 1; i < nregs; i++) {
                if (next_op->args[i] == srcreg
                    && (next_op->dead_regs & (1 << i))) {
                    /* The source value originally dies here.  Record this
                       and clear the death note.  */
                    src_dead = 1;
                    last_op = next_op;
                    next_op->dead_regs &= ~(1 << i);
                }
                if (next_op->args[i] == destreg) {
                    /* A use of the destination has been found.  */
                    if (next_op->dead_regs & (1 << i))
                        dest_dead = 1;

                    if (aggressive || dest_dead) {
                        /* Propagate the source value into this insn.  */
                        next_op->args[i] = srcreg;
                        last_op = next_op;
                    } else {
                        /* Something that probably isn't worth propagating.  */
                        break;
                    }
                }
            }
            /* If we decide not to replace a use, stop now.  */
            if (i < nregs)
                break;
            /* Check that the source value is not modified.  */
            if (next_op->args[0] == srcreg) {
                break;
            }
            if (dest_dead)
                break;
        }
        if (src_dead) {
            /* Mark the new death ocation for the source reg.  */
            nregs = qop_nregs[last_op->opcode - FIRST_QOP];
            for (i = 1; i < nregs; i++) {
                if (last_op->args[i] == srcreg) {
                    last_op->dead_regs |= (1 << i);
                }
            }
        }
        if (dest_dead) {
            /* This was the last use of this reg.
               Remove the original set.  */
            remove_qop(qop);
        }
    }
}
#endif

static void dump_all_ops(const char * msg)
{
    if (loglevel & (CPU_LOG_TB_OP_RAW)) {
        fprintf(logfile, msg);
        dump_ops(gen_first_qop);
        fprintf(logfile, "\n");
    }
}

static void dyngen_optimize(TranslationBlock *tb)
{
    dyngen_reglife();
    /*dyngen_cprop();*/
    dyngen_regalloc();
    dyngen_alloc_stack(tb);
    dump_all_ops("AFTER REGALLOC:\n");
}

static void dyngen_init_codegen()
{
    static int done_init = 0;
    if (!done_init) {
        /* ??? call this from somewhere else.  */
        dyngen_init_qop_info();
        /* Initialize target qregs.  */
        memcpy(gen_reginfo, target_reginfo,
               TARGET_NUM_REGS * sizeof(struct qreg));
        done_init = 1;
    }

    current_qop_buffer = &first_qop_buffer;
    qop_buffer_pos = 0;
    gen_free_qops = MAX_QOPS;
          
    gen_first_qop = NULL;
    gen_last_qop = NULL;
    nb_gen_labels = 0;

    HBM_CLEAR(spill_lru_mask);

    gen_last_qreg = TARGET_NUM_REGS;
}

/* return non zero if the very first instruction is invalid so that
   the virtual CPU can trigger an exception. 

   '*gen_code_size_ptr' contains the size of the generated code (host
   code).
*/
int cpu_gen_code(CPUState *env, TranslationBlock *tb,
                 int max_code_size, int *gen_code_size_ptr)
{
    uint8_t *gen_code_buf;
    int gen_code_size;

#ifdef USE_CODE_COPY
    if (code_copy_enabled &&
        cpu_gen_code_copy(env, tb, max_code_size, &gen_code_size) == 0) {
        /* nothing more to do */
    } else
#endif
    {
        dyngen_init_codegen();

        if (gen_intermediate_code(env, tb) < 0)
            return -1;

        dyngen_optimize(tb);

        /* generate machine code */
#ifdef USE_DIRECT_JUMP
        tb->tb_jmp_offset[0] = 0xffff;
        tb->tb_jmp_offset[1] = 0xffff;
#else
        tb->tb_next[0] = 0;
        tb->tb_next[1] = 0;
#endif
        gen_code_buf = tb->tc_ptr;

        dyngen_labels(gen_labels, nb_gen_labels, gen_code_buf, gen_first_qop);

        gen_code_size = dyngen_code(gen_code_buf,
#ifdef USE_DIRECT_JUMP
                                    tb->tb_jmp_offset,
#else
                                    NULL,
#endif
                                    gen_first_qop, gen_labels, gen_reginfo);
    }
    *gen_code_size_ptr = gen_code_size;
    /* ??? Could either make the label offsets 32bit, or
       dynamically recompile into smaller chunks.  */
    qemu_assert(gen_code_size < 0x10000, "Generated code too big");
#ifdef DEBUG_DISAS
    if (loglevel & CPU_LOG_TB_OUT_ASM) {
        fprintf(logfile, "OUT: [size=%d]\n", *gen_code_size_ptr);
        disas(logfile, tb->tc_ptr, *gen_code_size_ptr);
        fprintf(logfile, "\n");
        fflush(logfile);
    }
#endif
    return 0;
}

/* The cpu state corresponding to 'searched_pc' is restored. 
 */
int cpu_restore_state(TranslationBlock *tb, 
                      CPUState *env, unsigned long searched_pc,
                      void *puc)
{
    int c;
    qOP *qop;
    qOP *unwind;
    uint8_t *gen_code_ptr;

#ifdef USE_CODE_COPY
    if (tb->cflags & CF_CODE_COPY) {
        return cpu_restore_state_copy(tb, env, searched_pc, puc);
    }
#endif

    /* Regenerate the qops for this TB.  */
    dyngen_init_codegen();

    if (gen_intermediate_code(env, tb) < 0)
        return -1;

    dyngen_optimize(tb);

    /* Find the unwinding point immediately before the faulting
       instruction.  */
    gen_code_ptr = tb->tc_ptr;
    unwind = NULL;
    for (qop = gen_first_qop; qop; qop = qop->next) {
        if ((unsigned long)gen_code_ptr > searched_pc)
            break;

        c = qop->opcode;
        if (c == INDEX_op_unwind) {
            unwind = qop;
        } else if (c >= FIRST_QOP) {
            gen_code_ptr += size_qop(c,  qop->args, gen_reginfo);
        } else {
            gen_code_ptr += opc_copy_size[c];
        }
    }
    if (!unwind)
        cpu_abort(env, "Exception unwinding failed");

#if defined(TARGET_I386)
    {
        int cc_op;

        env->eip = tb->pc + unwind->args[0] - tb->cs_base;
        cc_op = unwind->args[1];
        if (cc_op != CC_OP_DYNAMIC)
            env->cc_op = cc_op;
    }
#elif defined(TARGET_ARM)
    env->regs[15] = tb->pc + unwind->args[0];
#elif defined(TARGET_SPARC)
    {
        int unwind_index;
        target_ulong npc;

        env->pc = tb->pc + unwind->args[0];
        unwind_index = unwind->args[1];
        npc = gen_unwind_info[unwind_index].npc;
        if (npc == 1) {
            /* dynamic NPC: already stored */
        } else if (npc == 2) {
            target_ulong t2 = (target_ulong)puc;
            /* jump PC: use T2 and the jump targets of the translation */
            if (t2) 
                env->npc = gen_unwind_info[unwind_index].jump_pc[0];
            else
                env->npc = gen_unwind_info[unwind_index].jump_pc[1];
        } else {
            env->npc = npc;
        }
    }
#elif defined(TARGET_PPC)
    {
        env->nip = tb->pc + unwind->args[0];
        env->access_type = unwind->args[1];
    }
#elif defined(TARGET_M68K)
    {
        int cc_op;
        env->pc = tb->pc + unwind->args[0];
        cc_op = unwind->args[1];
        if (cc_op != CC_OP_DYNAMIC)
            env->cc_op = cc_op;
    }
#elif defined(TARGET_MIPS)
    {
        env->PC = tb->pc + unwind->args[0];
        env->hflags &= ~MIPS_HFLAG_BMASK;
        env->hflags |= unwind->args[1];
    }
#else
#error unsupported CPU
#endif
    return 0;
}
